//==========================================================================
//#####DESCRIPTIONBEGIN####
//
// Author(s):    Atheros Communications, Inc.
// Contributors: Atheros Engineering
// Date:         2003-10-22
// Purpose:      
// Description:  AR531X ethernet hardware driver
//              
//
//####DESCRIPTIONEND####
//
//==========================================================================
/*
 * Manage the ethernet PHY.
 * This code supports ADMTek Phy. The ADK Tek on power up is already configured 
 * to auto negotiate. Hence auto negotiation seetings are not configured again.
 * AUTO MDIX is one feature which is not enabled by default. This is configured
 * in software.
 */

#if defined(linux)
#include <linux/config.h>
#include <linux/types.h>
#include <linux/netdevice.h>
#include <linux/etherdevice.h>
#include <linux/delay.h>

#include "ar531xlnx.h"
#endif

extern UINT16 ae531x_MiiRead(UINT32 phyBase, UINT32 phyAddr, UINT8 reg);
extern void ae531x_MiiWrite(UINT32 phyBase, UINT32 phyAddr, UINT8 reg, UINT16 data);

#define phyRegRead ae531x_MiiRead
#define phyRegWrite ae531x_MiiWrite

#if /*DEBUG */ 1
#define ADM_DEBUG_ERROR     0x00000001
#define ADM_DEBUG_PHYSETUP  0x00000002
#define ADM_DEBUG_PHYCHANGE 0x00000004

int rtPhyDebug = ADM_DEBUG_ERROR ;

#define ADM_PRINT(FLG, X)                            \
{                                                   \
    if (rtPhyDebug & (FLG)) {                       \
        DEBUG_PRINTF X;                             \
    }                                               \
}
#else
#define ADM_PRINT(FLG, X)
#endif


/* Convert from phy unit# to (phyBase, phyAddr) pair */
static UINT32 admPhyBase;

#define ADM_PHYBASE(phyUnit) (admPhyBase)


#define ADM_CHIP_ID1_EXPECTATION                   0x1020 
#define ADM_CHIP_ID2_EXPECTATION                   0x0007 
#define ADM_PHY_ADDR                               0x5
#define PHY_ADDR_SW_PORT 0

void setupAdmTekAutoMdix(void) {
#define PHY_ADDR_SW_PORT 0
#define ADM_SW_AUTO_MDIX_EN     0x8000
   UINT32  phyUnit;
   UINT16  reg = 0;
   UINT32  phyBase;
   /* sw_port address for ports 0-5. Found on page 4-1 of ADM6996FC datasheet v1.0 */   
   UINT32 sw_port_addr[6] = {1,3,5,7,8,9};

   phyBase = admPhyBase;
   
   for(phyUnit=0;phyUnit<6;phyUnit++) {
      reg = phyRegRead(phyBase,PHY_ADDR_SW_PORT,sw_port_addr[phyUnit]);
      ADM_PRINT(ADM_DEBUG_PHYSETUP,("reg(0x%x)=0x%x\n",sw_port_addr[phyUnit],reg));
      reg |= ADM_SW_AUTO_MDIX_EN;
      phyRegWrite(phyBase,PHY_ADDR_SW_PORT,sw_port_addr[phyUnit],reg);
   }

}

#ifdef CONFIG_VENETDEV
#define LAN_PORT(x)			(x < 4)
#define WAN_PORT(x)			(x == 4)
#define ADM_VLAN_MODE_SELECT_REG	0x11
#define ADM_VLAN_MODE_VM_S		5
#define ADM_VLAN_MODE_8021Q		0x1
#define ADM_VLAN_MAC_CLONE_S		4
#define ADM_VLAN_MAC_CLONE_EN		0x1
#define ADM_VLAN_MAP_REG_1		0x14
#define ADM_VLAN_MAP_REG_2		0x15
#define ADM_VLAN_MAP_PORT_0		0x0001
#define ADM_VLAN_MAP_PORT_1		0x0004
#define ADM_VLAN_MAP_PORT_2		0x0010
#define ADM_VLAN_MAP_PORT_3		0x0040
#define ADM_VLAN_MAP_PORT_4		0x0080
#define ADM_VLAN_MAP_PORT_5		0x0100
void admVLANInit(void)
{
    UINT32 phyUnit;
    UINT16 reg = 0;
    UINT32 phyBase = admPhyBase;
    UINT32 sw_port_addr[6] = {1,3,5,7,8,9};
    
    for(phyUnit = 0; phyUnit < 6; phyUnit++)
    {
	reg = phyRegRead(phyBase,PHY_ADDR_SW_PORT,sw_port_addr[phyUnit]);
	if (LAN_PORT(phyUnit))
	    reg = 0x840f;
	else if (WAN_PORT(phyUnit))
	    reg = 0x880f;
	else /* CPU port */
	    reg = 0x881f;
	phyRegWrite(phyBase, PHY_ADDR_SW_PORT, sw_port_addr[phyUnit], reg);
    }
    reg = phyRegRead(phyBase, PHY_ADDR_SW_PORT, ADM_VLAN_MODE_SELECT_REG);
    reg |= (ADM_VLAN_MODE_8021Q << ADM_VLAN_MODE_VM_S);
    reg |= (ADM_VLAN_MAC_CLONE_EN << ADM_VLAN_MAC_CLONE_S);
    phyRegWrite(phyBase, PHY_ADDR_SW_PORT, ADM_VLAN_MODE_SELECT_REG, reg);

    reg = (ADM_VLAN_MAP_PORT_0 | ADM_VLAN_MAP_PORT_1 | ADM_VLAN_MAP_PORT_2 | 
        ADM_VLAN_MAP_PORT_3 | ADM_VLAN_MAP_PORT_5);
    phyRegWrite(phyBase, PHY_ADDR_SW_PORT, ADM_VLAN_MAP_REG_1, reg);
    reg = (ADM_VLAN_MAP_PORT_4 | ADM_VLAN_MAP_PORT_5);
    phyRegWrite(phyBase, PHY_ADDR_SW_PORT, ADM_VLAN_MAP_REG_2, reg);
}
#endif

/******************************************************************************
*
* phySetup - reset and setup the PHY associated with
* the specified MAC unit number.
*
* Resets the associated PHY port.
*
* RETURNS:
*    TRUE  --> associated PHY is alive
*    FALSE --> no LINKs on this ethernet unit
*/

BOOL
phySetup(int ethUnit, UINT32 phyBase)
{

    UINT16  phyID1;
    UINT16  phyID2;

    admPhyBase = phyBase;

    /* Checking whether this is ADMTek Phy. This is used on AP61. */

    phyID1 = phyRegRead(phyBase, 0x5, 0x0);
    phyID2 = phyRegRead(phyBase, 0x5, 0x1);
    ADM_PRINT(ADM_DEBUG_PHYSETUP,("phyID1=0x%x phyId2=0x%x.\n",phyID1,phyID2));
    if(((phyID1 & 0xfff0) == ADM_CHIP_ID1_EXPECTATION) && (phyID2 == ADM_CHIP_ID2_EXPECTATION)){
       ADM_PRINT(ADM_DEBUG_PHYSETUP,("Found ADM6996FC.phyID1=0x%x phyId2=0x%x.\n",phyID1,phyID2));
#ifdef CONFIG_VENETDEV
       admVLANInit();
#else
       setupAdmTekAutoMdix();
#endif
    }
    return TRUE;

}



/******************************************************************************
*
* phyIsDuplexFull - Determines whether the phy ports associated with the
* specified device are FULL or HALF duplex.
*
* RETURNS:
*    TRUE  --> FULL
*    FALSE --> HALF
*/
BOOL
phyIsFullDuplex(int ethUnit)
{

   ethUnit = 0; /* Keep compiler happy*/
   return TRUE;

}

/******************************************************************************
*
* rt_phyIsSpeed100 - Determines the speed of phy ports associated with the
* specified device.
*
* RETURNS:
*    TRUE --> 100Mbit
*    FALSE --> 10Mbit
*/
BOOL
phyIsSpeed100(int phyUnit)
{
   phyUnit = 0; /* Keep compiler happy*/
   return TRUE;
}

/*****************************************************************************
*
* phyCheckStatusChange -- checks for significant changes in PHY state.
*
* A "significant change" is:
*     dropped link (e.g. ethernet cable unplugged) OR
*     autonegotiation completed + link (e.g. ethernet cable plugged in)
*
*/
void
phyCheckStatusChange(int ethUnit)
{
   ethUnit = 0; /* Keep compiler happy*/
}

/* foxconn added start, wenchia, 01/12/2007 */
int phyGetPortStatus(int portNum)
{
    UINT16 portStatus;
    UINT32 phyBase;
    UINT32 retVal;

    phyBase = admPhyBase;

    switch (portNum)
    {
        case 0:
            portStatus = phyRegRead(phyBase, 0x5, 0x2);
            retVal = portStatus & 0x0000000f; //p0
            break;
        case 1:
            portStatus = phyRegRead(phyBase, 0x5, 0x2);
            retVal = (portStatus>>8) & 0x0000000f; //p1
            break;
        case 2:
            portStatus = phyRegRead(phyBase, 0x5, 0x3);
            retVal = portStatus & 0x0000000f; //p2
            break;
        case 3:
            portStatus = phyRegRead(phyBase, 0x5, 0x3);
            retVal = (portStatus>>8) & 0x0000000f; //p3
            break;
        case 4:
            portStatus = phyRegRead(phyBase, 0x5, 0x3);
            retVal = (portStatus>>12) & 0x0000000f; //p4
            break;
        case 5:
            portStatus = phyRegRead(phyBase, 0x5, 0x4);
            retVal = ((portStatus>>1)&0x0000000c) | (portStatus&0x00000003);//p5
            break;
        default:
            retVal = 0;
            break;
    }
    return retVal;
}

void phySetLanPwrDown(int pwrDown)
{
    UINT16  reg = 0;
    UINT32 phyBase;
    
    phyBase = admPhyBase;
    if (pwrDown) {
        reg = phyRegRead(phyBase, 0x10, 0x0); /*PHY_C0, offset=0x200*/
        reg |= 0x0800; /* power down */
        phyRegWrite(phyBase, 0x10, 0x0, reg);
        
        reg = phyRegRead(phyBase, 0x11, 0x0); /*PHY_C1, offset=0x220*/
        reg |= 0x0800; /* power down */
        phyRegWrite(phyBase, 0x11, 0x0, reg);
        
        reg = phyRegRead(phyBase, 0x12, 0x0); /*PHY_C2, offset=0x240*/
        reg |= 0x0800; /* power down */
        phyRegWrite(phyBase, 0x12, 0x0, reg);
        
        reg = phyRegRead(phyBase, 0x13, 0x0); /*PHY_C3, offset=0x260*/
        reg |= 0x0800; /* power down */
        phyRegWrite(phyBase, 0x13, 0x0, reg);
    } else {
        reg = phyRegRead(phyBase, 0x10, 0x0); /*PHY_C0, offset=0x200*/
        reg &= 0xf7ff; /* normal operation */
        phyRegWrite(phyBase, 0x10, 0x0, reg);
        
        reg = phyRegRead(phyBase, 0x11, 0x0); /*PHY_C1, offset=0x220*/
        reg &= 0xf7ff; /* normal operation */
        phyRegWrite(phyBase, 0x11, 0x0, reg);
        
        reg = phyRegRead(phyBase, 0x12, 0x0); /*PHY_C2, offset=0x240*/
        reg &= 0xf7ff; /* normal operation */
        phyRegWrite(phyBase, 0x12, 0x0, reg);
        
        reg = phyRegRead(phyBase, 0x13, 0x0); /*PHY_C3, offset=0x260*/
        reg &= 0xf7ff; /* normal operation */
        phyRegWrite(phyBase, 0x13, 0x0, reg);
    }
}

int phyGetResetButtonStatus(void)
{
    if (sysRegRead(AR531XPLUS_GPIO_DI) & (1 << 6)) //gpio 6 (reset button)
    {
        //reset button is pressed
        return 1;
    }
    else
    {
        //reset button is released
        return 0;
    }
}
/* foxconn added end, wenchia, 01/12/2007 */
