/*
 * ########################################################################
 *
 *  This program is free software; you can redistribute  it and/or modify it
 *  under  the terms of  the GNU General  Public License as published by the
 *  Free Software Foundation;  either version 2 of the	License, or (at your
 *  option) any later version.
 *
 *  THIS  SOFTWARE  IS PROVIDED   ``AS	IS'' AND   ANY	EXPRESS OR IMPLIED
 *  WARRANTIES,   INCLUDING, BUT NOT  LIMITED  TO, THE IMPLIED WARRANTIES OF
 *  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
 *  NO	EVENT  SHALL   THE AUTHOR  BE	 LIABLE FOR ANY   DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 *  NOT LIMITED   TO, PROCUREMENT OF  SUBSTITUTE GOODS	OR SERVICES; LOSS OF
 *  USE, DATA,	OR PROFITS; OR	BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 *  ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 *  THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *  You should have received a copy of the  GNU General Public License along
 *  with this program; if not, write  to the Free Software Foundation, Inc.,
 *  675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * ########################################################################
*************************************************************************************
* acacia.c
* driver for the IDT RC32438 on-chip ethernet controller.
*************************************************************************************
Author: MontaVista Software, Inc.
stevel@mvista.com or source@mvista.com

Author: Integrated Device Technology, Inc.
Maruthanayagam: 	Ported from banyan.c 
cgg:                    	Modified code.
*************************************************************************************
Haofeng Kou			6/7/03
Using the IOD COD and DMA chaining mode to handle the tansmit
In order to prevent the DMA Race Condition, make software access NDPTR only for tansmit.
Using the IOF COF and NDPTR of DMA chaining to handle the Receive.
In order to prevent the DMA Race Condition, use NDPTR trigger DMA chaining once
overwrite happen for receive.
Modify acacia_restart() to make it recover Bridge mode setting after reset.
Adjust the dev_kfree_skb_any() and dev_kfree_skb_irq() usage.
Adjust the netif_wake_queue() and netif_start_queue() usage.
Add in code to handle the case once there is no memory in system(only for special situations).
*************************************************************************************
Haofeng Kou			6/16/03
Add in code to handle the TX Underflow.
Trigger the waiting queue in the TX ISR (7/23/03)
Fix the bug happen for the wu-ftp server (9/15/03)
*************************************************************************************
P. Sadik  Oct 10, 2003
'Management Clock Prescaler Divisor' is no longer determined at compile time, but
is caculated at run time, based on idt_cpu_freq.
*************************************************************************************
Haofeng Kou			10/27/03
Change the ISR to use Tasklet for the TX and RX.
Add in the DMA Abort code and use DMA Abort instead of the DMA halt. (DMA abort finished 
faster than DMA halt, so replace the halt code by abort)
Optimize the interrupt disable/enable actions to reduce the number of interrupt.
Improve the CPU usage.
*************************************************************************************
P. Sadik Nov 14, 2003
For Acacia ZC part, the receive handler is modified such that, the DMA halt
condition is checked after the taskelet has finished handling all descriptors.
If the H bit is set, then the DMA is restarted. With this fix, the overflow
interrupt need not be taken. 
**********************************************************************************
Haofeng Kou			11/18/03
Merge together the ZA and ZC code, auto-switch between the ZA and ZC.
Update all the debug features(ASSERT, DBG, ERR, INFO and WARN).
Add in the loadable driver module support.
Add the MII-PHY LXT973 duplex mode auto-negotiation support.(Not ebable at this time)
*************************************************************************************
rkt                            4/29/04
Added Fast-Switching feature.
acacia_rx_tasklet routine modified to handle race conditions in the driver
*************************************************************************************

*/

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/ctype.h>
#include <linux/types.h>
#include <linux/fcntl.h>
#include <linux/interrupt.h>
#include <linux/ptrace.h>
#include <linux/init.h>
#include <linux/ioport.h>
#include <linux/proc_fs.h>
#include <linux/in.h>
#include <linux/slab.h> 
#include <linux/string.h>
#include <linux/delay.h>
#include <linux/netdevice.h>
#include <linux/etherdevice.h>
#include <linux/skbuff.h>
#include <linux/errno.h>
#include <linux/interrupt.h>
#include <asm/bootinfo.h>
#include <asm/system.h>
#include <asm/bitops.h>
#include <asm/pgtable.h>
#include <asm/segment.h>
#include <asm/io.h>
#include <asm/dma.h>
#include <asm/atomic.h>

#ifdef CONFIG_NET_FASTROUTE
#include <linux/if_arp.h>
#include <net/ip.h>
#endif

#include "acacia.h"
#include  <asm/rc32438/rc32438.h>



#define STATION_ADDRESS_HIGH(dev) (((dev)->dev_addr[0] << 8) | \
			           ((dev)->dev_addr[1]))
#define STATION_ADDRESS_LOW(dev)  (((dev)->dev_addr[2] << 24) | \
				   ((dev)->dev_addr[3] << 16) | \
				   ((dev)->dev_addr[4] << 8)  | \
				   ((dev)->dev_addr[5]))

#define MII_CLOCK 1250000 				/* no more than 2.5MHz */
static char mac0[7] = { 0x08, 0x00, 0x06, 0x05, 0x40, 0x01 }; 
static char mac1[7] = { 0x08, 0x00, 0x06, 0x05, 0x50, 0x01 }; 

static struct idt32438_boarddata *idt32438_boardConfig=NULL;


MODULE_PARM(mac0, "c7");
MODULE_PARM(mac1, "c7");
MODULE_PARM_DESC(mac0, "MAC address for RC32438 ethernet0");
MODULE_PARM_DESC(mac1, "MAC address for RC32438 ethernet1");

static struct acacia_if_t {
	struct net_device *dev;
	char* mac_str;
	u32 iobase;
	int rx_dma_irq;
	int tx_dma_irq;
	int rx_ovr_irq;
  	int tx_und_irq;			
} acacia_iflist[] = 
	{
	 	{	NULL, mac0, ETH1_PhysicalAddress,
	 		ETH1_DMA_RX_IRQ, ETH1_DMA_TX_IRQ, ETH1_RX_OVR_IRQ, ETH1_TX_UND_IRQ},
		{	NULL, mac1, ETH0_PhysicalAddress,
	 		ETH0_DMA_RX_IRQ, ETH0_DMA_TX_IRQ, ETH0_RX_OVR_IRQ, ETH0_TX_UND_IRQ},
		{NULL, 0, 0}
	};

#if 0
static int parse_mac_addr(struct net_device *dev, char* macstr)
{
	int i, j;
	unsigned char result, value;
		
	for (i=0; i<6; i++) 
	{
		result = 0;
		if (i != 5 && *(macstr+2) != ':') 
		{
			ERR(__FILE__ "invalid mac address format: %d %c\n",
			    i, *(macstr+2));
			return -EINVAL;
		}				
		for (j=0; j<2; j++) 
		{
			if (isxdigit(*macstr) && (value = isdigit(*macstr) ? *macstr-'0' : 
			     toupper(*macstr)-'A'+10) < 16) 
			{
				result = result*16 + value;
				macstr++;
			} 
			else 
			{
				ERR(__FILE__ "invalid mac address "
				    "character: %c\n", *macstr);
				return -EINVAL;
			}
		}
				
		macstr++; 
		dev->dev_addr[i] = result;
	}
		
	return 0;
}

#endif

static inline void acacia_abort_tx(struct net_device *dev)
{
	struct acacia_local *lp = (struct acacia_local *)dev->priv;

	rc32438_abort_dma(dev, lp->tx_dma_regs);
	
}
static inline void acacia_abort_rx(struct net_device *dev)
{
	struct acacia_local *lp = (struct acacia_local *)dev->priv;

	rc32438_abort_dma(dev, lp->rx_dma_regs);
	
}
static inline void acacia_halt_tx(struct net_device *dev)
{
	struct acacia_local *lp = (struct acacia_local *)dev->priv;
	
	if (rc32438_halt_dma(lp->tx_dma_regs))
		ERR(__FUNCTION__ ": timeout!\n");
}
static inline void acacia_halt_rx(struct net_device *dev)
{
	struct acacia_local *lp = (struct acacia_local *)dev->priv;

	if (rc32438_halt_dma(lp->rx_dma_regs))
		ERR(__FUNCTION__ ": timeout!\n");
}
static inline void acacia_start_tx(struct acacia_local *lp,  volatile DMAD_t td)
{
	rc32438_start_dma(lp->tx_dma_regs, virt_to_phys(td));
}
static inline void acacia_start_rx(struct acacia_local *lp, volatile DMAD_t rd)
{
	rc32438_start_dma(lp->rx_dma_regs, virt_to_phys(rd));
}

static inline void acacia_chain_tx(struct acacia_local *lp, volatile DMAD_t td)
{
	rc32438_chain_dma(lp->tx_dma_regs, virt_to_phys(td));
}
static inline void acacia_chain_rx(struct acacia_local *lp, volatile DMAD_t rd)
{
	rc32438_chain_dma(lp->rx_dma_regs, virt_to_phys(rd));
}

#ifdef ACACIA_PROC_DEBUG
static int acacia_read_proc(char *buf, char **start, off_t fpos,
			    int length, int *eof, void *data)
{
	struct net_device *dev = (struct net_device *)data;
	struct acacia_local *lp = (struct acacia_local *)dev->priv;
	int len = 0;

	/* print out header */
	len += sprintf(buf + len, "\n\tRC32438 Ethernet Debug\n\n");
	len += sprintf (buf + len,
			"DMA halt count      = %10d, total pkt cnt = %10d\n",
			lp->dma_halt_cnt, lp->halt_tx_count);
	len += sprintf (buf + len,
			"DMA run count       = %10d, total pkt cnt = %10d\n",
			lp->dma_run_cnt, lp->run_tx_count);
	len += sprintf (buf + len,
			"DMA race count      = %10d, total pkt cnt = %10d\n",
			lp->dma_race_cnt, lp->race_tx_count);
	len += sprintf (buf + len,
			"DMA collision count = %10d, total pkt cnt = %10d\n",
			lp->dma_collide_cnt, lp->collide_tx_count);

	if (fpos >= len) 
	{
		*start = buf;
		*eof = 1;
		return 0;
	}
	*start = buf + fpos;
	
	if ((len -= fpos) > length)
		return length;	
	*eof = 1;
	
	return len;

}
#endif


#ifdef CONFIG_NET_FASTROUTE	

static int acacia_eth_accept_fastpath(struct net_device *dev, struct dst_entry *dst)
{
	struct net_device *odev = dst->dev;

	if ((dst->ops->protocol != __constant_htons(ETH_P_IP))
	    || (odev->type != ARPHRD_ETHER)
	    || !odev->accept_fastpath) {
		return (-1);
	}
	return (0);
}
#endif


int acacia_get_board_config(void)
{
    int dataFound;
    char *bd_config;

    /*
     * Find start of Board Configuration data, using heuristics:
     * Search back from the (aliased) end of flash by 0x1000 bytes
     * at a time until we find the string "PB32", which marks the
     * start of Board Configuration.  Give up if we've searched
     * more than 500KB.
     */
    dataFound = 0;
    for (bd_config = (char *)0xbffff000;
         bd_config > (char *)0xbff80000;
         bd_config -= 0x1000)
    {
        if ( *(int *)bd_config == IDT32438_BD_MAGIC) {
            dataFound = 1;
            break;
        }
    }

    if (!dataFound) {
        printk("Could not find Board Configuration Data\n");
                bd_config = NULL;
    }

    idt32438_boardConfig = (struct idt32438_boarddata *) bd_config;

    return(dataFound);
}

/*
 * Fetch a pointer to an ethernet's MAC address
 * in the Board Configuration data (in flash).
 */
char *
acacia_enet_mac_address_get(int MACUnit)
{
        /* XXX: Hack for poorly configured boards.
         *      Cannot setup bridging properly (brctl) when both enet
         *      interfaces share the same MAC address.
         *
         */

#ifdef CONFIG_ASK_MULT_MAC_HACK
    static u8  enet0Mac[6] = {0x00, 0x0d, 0x0b, 0x13, 0x6b, 0x16};
    static u8  enet1Mac[6] = {0x00, 0x0d, 0x0b, 0x13, 0x6b, 0x17};
#endif

    if (MACUnit == 0) {
#ifndef CONFIG_ASK_MULT_MAC_HACK
        return idt32438_boardConfig->enet0Mac;
#else
                return enet0Mac;
#endif
    }
    if (MACUnit == 1) {
#ifndef CONFIG_ASK_MULT_MAC_HACK
        return idt32438_boardConfig->enet1Mac;
#else
                return enet1Mac;
#endif
    }
    printk("Invalid ethernet MAC unit number (%d)!\n", MACUnit);
    return NULL;
}


/*
 * Restart the ACACIA ethernet controller. 
 */
static int acacia_restart(struct net_device *dev)
{
	struct acacia_local *lp = (struct acacia_local *)dev->priv;
	
	/*
	 * Disable interrupts
	 */
	disable_irq(lp->rx_irq);
	disable_irq(lp->tx_irq);
#ifdef	RC32438_REVISION_ZA
	disable_irq(lp->ovr_irq);
#endif	
	disable_irq(lp->und_irq);

	/* Mask F E bit in Tx DMA */
	local_writel(local_readl(&lp->tx_dma_regs->dmasm) | DMASM_f_m | DMASM_e_m, &lp->tx_dma_regs->dmasm);
	/* Mask D H E bit in Rx DMA */
	local_writel(local_readl(&lp->rx_dma_regs->dmasm) | DMASM_d_m | DMASM_h_m | DMASM_e_m, &lp->rx_dma_regs->dmasm);

	acacia_init(dev);
	acacia_multicast_list(dev);

	enable_irq(lp->und_irq);
#ifdef	RC32438_REVISION_ZA
	enable_irq(lp->ovr_irq);
#endif
	enable_irq(lp->tx_irq);
	enable_irq(lp->rx_irq);

	return 0;
}

int acacia_init_module(void)
{
	int i, retval=0;

	for (i = 0; acacia_iflist[i].iobase; i++) 
	{ 
		retval |= acacia_probe(i);
	} 

	return retval;
}

static int acacia_probe(int port_num)
{
	struct acacia_local *lp = NULL;
	struct acacia_if_t *bif = &acacia_iflist[port_num];
	struct net_device *dev = NULL;
	int i, retval;
	char *addr;
	
	request_region(acacia_iflist[port_num].iobase, 0x24C, "ACACIA");
		
	/* Allocate a new 'dev' if needed */
	dev = init_etherdev(0, sizeof(struct acacia_local));

	bif->dev = dev;
	
	INFO("RC32438 ethernet%d found at 0x%08x\n", 
		port_num, acacia_iflist[port_num].iobase);

	/* Fill in the 'dev' fields. */
	dev->base_addr = bif->iobase;
	/* just use the rx dma irq */
	dev->irq = bif->rx_dma_irq; 

	if (acacia_get_board_config()) {
		/* Board Config data exists. Get it */
	        addr = acacia_enet_mac_address_get(port_num);
		/* if MAC is bogus in config data, skip */
        	if((*(u32 *)addr == 0xffffffff) && (*(u16 *)(addr+4)==0xffff)){
			printk("ERROR: Bogus MAC config Data\n");
			printk("Using Default MAC address \n");
			addr = bif->mac_str;
        	}
    	}
	else {
		printk("Using Default MAC address \n");
		addr = bif->mac_str;
	}
    	memcpy(dev->dev_addr, addr, dev->addr_len );

	INFO("HW Address ");
	for (i = 0; i < 6; i++) 
	{
		printk("%2.2x", dev->dev_addr[i]);
		if (i<5)
			printk(":");
	}
	printk("\n");
    
	INFO("Rx IRQ %d, Tx IRQ %d\n", bif->rx_dma_irq, bif->tx_dma_irq);
    
	/* Initialize the device structure. */
	if (dev->priv == NULL) 
	{
		lp = (struct acacia_local *)kmalloc(sizeof(*lp), GFP_KERNEL);
		memset(lp, 0, sizeof(struct acacia_local));
	} 
	else 
	{
		lp = (struct acacia_local *)dev->priv;
	}

	dev->priv = lp;
	lp->rx_irq = bif->rx_dma_irq;
	lp->tx_irq = bif->tx_dma_irq;
	lp->ovr_irq = bif->rx_ovr_irq;
	lp->und_irq = bif->tx_und_irq;

	lp->eth_regs = ioremap_nocache(bif->iobase,
				       sizeof(*lp->eth_regs));

	if (!lp->eth_regs) 
	{
		ERR("Can't remap eth registers\n");
		retval = -ENXIO;
		goto probe_err_out;
	}

	if (port_num == 0)
	{
	    	lp->rx_dma_regs =
			ioremap_nocache(DMA0_PhysicalAddress + 4 * DMA_CHAN_OFFSET,	sizeof(struct DMA_Chan_s));

	     	if (!lp->rx_dma_regs) 
		{
			ERR("Can't remap Rx DMA registers\n");
			retval = -ENXIO;
			goto probe_err_out;
		}		
	    	lp->tx_dma_regs =
			ioremap_nocache(DMA0_PhysicalAddress  + 5 * DMA_CHAN_OFFSET,
				sizeof(struct DMA_Chan_s));

	     	if (!lp->tx_dma_regs) 
		{
			ERR("Can't remap Tx DMA registers\n");
			retval = -ENXIO;
			goto probe_err_out;
		}

#ifdef ACACIA_PROC_DEBUG
	lp->ps = create_proc_read_entry ("net/rc32438_0", 0, NULL,
					 acacia_read_proc, dev);
#endif	
	}
     	else if (port_num == 1)
	{
		lp->rx_dma_regs =
			ioremap_nocache(DMA0_PhysicalAddress  + 2 * DMA_CHAN_OFFSET,
				sizeof(struct DMA_Chan_s));

	     	if (!lp->rx_dma_regs) 
		{
			ERR("Can't remap Rx DMA registers\n");
			retval = -ENXIO;
			goto probe_err_out;
	     	}
		
	    	lp->tx_dma_regs =
		ioremap_nocache(DMA0_PhysicalAddress  + 3 * DMA_CHAN_OFFSET,
				sizeof(struct DMA_Chan_s));

	     	if (!lp->tx_dma_regs) 
		{
			ERR("Can't remap Tx DMA registers\n");
			retval = -ENXIO;
			goto probe_err_out;
	     	}
#ifdef ACACIA_PROC_DEBUG
	lp->ps = create_proc_read_entry ("net/rc32438_1", 0, NULL,
					 acacia_read_proc, dev);
#endif
	   
     	}

	lp->td_ring = (DMAD_t)kmalloc(TD_RING_SIZE + RD_RING_SIZE, GFP_KERNEL);
	if (!lp->td_ring) 
	{
		ERR("Can't allocate descriptors\n");
		retval = -ENOMEM;
		goto probe_err_out;
	}

	dma_cache_inv((unsigned long)(lp->td_ring),TD_RING_SIZE + RD_RING_SIZE);

	/* now convert TD_RING pointer to KSEG1 */
	//lp->td_ring = (DMAD_t )KSEG1ADDR(lp->td_ring);

	lp->rd_ring = &lp->td_ring[ACACIA_NUM_TDS];

	spin_lock_init(&lp->lock);

	dev->open = acacia_open;
	dev->stop = acacia_close;
	dev->hard_start_xmit = acacia_send_packet;

#ifdef CONFIG_NET_FASTROUTE
	dev->accept_fastpath = acacia_eth_accept_fastpath;
#endif
	dev->get_stats	= acacia_get_stats;
	dev->set_multicast_list = &acacia_multicast_list;
	dev->tx_timeout = acacia_tx_timeout;
	dev->watchdog_timeo = ACACIA_TX_TIMEOUT;

	lp->rx_tasklet = kmalloc(sizeof(struct tasklet_struct), GFP_KERNEL);
	tasklet_init(lp->rx_tasklet, acacia_rx_tasklet, (unsigned long)dev);
	lp->tx_tasklet = kmalloc(sizeof(struct tasklet_struct), GFP_KERNEL);
	tasklet_init(lp->tx_tasklet, acacia_tx_tasklet, (unsigned long)dev);

	/* Fill in the fields of the device structure with ethernet values. */
	ether_setup(dev);
	return 0;

 probe_err_out:
	acacia_cleanup_module();
	ERR(__FUNCTION__ " failed.  Returns %d\n", retval);
	return retval;
	
}


/*
 * Open/initialize the ACACIA controller.
 *
 * This routine should set everything up anew at each open, even
 *  registers that "should" only need to be set once at boot, so that
 *  there is non-reboot way to recover if something goes wrong.
 */

static int acacia_open(struct net_device *dev)
{
	struct acacia_local *lp = (struct acacia_local *)dev->priv;

	MOD_INC_USE_COUNT;
	
	/* Initialize */
	if (acacia_init(dev)) 
	{
	  	ERR("Erroe: cannot open the Ethernet device\n");
	 	return -EAGAIN;
	}
	
	/* Install the interrupt handler that handles the Done Finished Ovr and Und Events */	
	if (request_irq(lp->rx_irq, &acacia_rx_dma_interrupt,
			SA_SHIRQ | SA_INTERRUPT,
			"acacia ethernet Rx", dev)) 
	{
		ERR(__FUNCTION__ ": unable to get Rx DMA IRQ %d\n",
		    lp->rx_irq);
		MOD_DEC_USE_COUNT;
		return -EAGAIN;
	}
	if (request_irq(lp->tx_irq, &acacia_tx_dma_interrupt,
			SA_SHIRQ | SA_INTERRUPT,
			"acacia ethernet Tx", dev)) 
	{
		ERR(__FUNCTION__ ": unable to get Tx DMA IRQ %d\n",
		    lp->tx_irq);
		free_irq(lp->rx_irq, dev);
		MOD_DEC_USE_COUNT;
		return -EAGAIN;
	}
	
#ifdef	RC32438_REVISION_ZA
	/* Install handler for overrun error. */
	if (request_irq(lp->ovr_irq, &acacia_ovr_interrupt,
			SA_SHIRQ | SA_INTERRUPT,
			"Ethernet Overflow", dev)) 
	{
		ERR(__FUNCTION__ ": unable to get OVR IRQ %d\n",
		    lp->ovr_irq);
		free_irq(lp->rx_irq, dev);
		free_irq(lp->tx_irq, dev);
		MOD_DEC_USE_COUNT;
		return -EAGAIN;
	}
#endif

	/* Install handler for underflow error. */
	if (request_irq(lp->und_irq, &acacia_und_interrupt,
			SA_SHIRQ | SA_INTERRUPT,
			"Ethernet Underflow", dev)) 
	{
		ERR(__FUNCTION__ ": unable to get UND IRQ %d\n",
		    lp->und_irq);
		free_irq(lp->rx_irq, dev);
		free_irq(lp->tx_irq, dev);
#ifdef	RC32438_REVISION_ZA		
		free_irq(lp->ovr_irq, dev);		
#endif
		MOD_DEC_USE_COUNT;
		return -EAGAIN;
	}

	/*Start MII-PHY Timer*/
	//Not enabled this feature at this time.
	/*
	init_timer(&lp->mii_phy_timer);
   	lp->mii_phy_timer.expires = jiffies + 10 * HZ;	
   	lp->mii_phy_timer.data = (unsigned long)dev;
   	lp->mii_phy_timer.function	 = acacia_mii_handler;
   	add_timer(&lp->mii_phy_timer);
	*/
	
	return 0;
}



/*
 * Close the ACACIA device
 */
static int acacia_close(struct net_device *dev)
{
	struct acacia_local *lp = (struct acacia_local *)dev->priv;
	u32 tmp;
    
	/* Disable interrupts */
	disable_irq(lp->rx_irq);
	disable_irq(lp->tx_irq);
#ifdef	RC32438_REVISION_ZA
	disable_irq(lp->ovr_irq);
#endif
	disable_irq(lp->und_irq);
		
	tmp = local_readl(&lp->tx_dma_regs->dmasm);
	tmp = tmp | DMASM_f_m | DMASM_e_m;
	local_writel(tmp, &lp->tx_dma_regs->dmasm);
     
	tmp = local_readl(&lp->rx_dma_regs->dmasm);
	tmp = tmp | DMASM_d_m | DMASM_h_m | DMASM_e_m;
	local_writel(tmp, &lp->rx_dma_regs->dmasm);

	free_irq(lp->rx_irq, dev);
	free_irq(lp->tx_irq, dev);
#ifdef	RC32438_REVISION_ZA	
	free_irq(lp->ovr_irq, dev);
#endif
	free_irq(lp->und_irq, dev);

	//Not enabled this feature at this time.
	//del_timer(&lp->mii_phy_timer);

	MOD_DEC_USE_COUNT;
	return 0;
}


/* transmit packet */
static int acacia_send_packet(struct sk_buff *skb, struct net_device *dev)
{
 	struct acacia_local		*lp = (struct acacia_local *)dev->priv;
  	unsigned long 			flags;
  	u32					length;
 	DMAD_t				td;


	spin_lock_irqsave(&lp->lock, flags);

	td = &lp->td_ring[lp->tx_chain_tail];

 	/* stop queue when full, drop pkts if queue already full */
  	if(lp->tx_count >= (ACACIA_NUM_TDS - 2))
  	{
		lp->tx_full = 1;
		
	  	if(lp->tx_count == (ACACIA_NUM_TDS - 2))
	  	{
			/* this pkt is about to fill the queue*/
			ERR("Tx Ring now full, queue stopped.\n");	
			netif_stop_queue(dev);
	  	}
		else
		{
			/* this pkt cannot be added to the full queue */
			ERR("Tx ring full, packet dropped\n");
			lp->stats.tx_dropped++;
			dev_kfree_skb_any(skb);
			spin_unlock_irqrestore(&lp->lock, flags);
			return 1;
	  	}	   
  	}	 

  	lp->tx_count ++;
	
 	if (lp->tx_skb[lp->tx_chain_tail] != NULL)
  		dev_kfree_skb_any(lp->tx_skb[lp->tx_chain_tail]);


	lp->tx_skb[lp->tx_chain_tail] = skb;
  
  	length = skb->len;

  	/* Setup the transmit descriptor. */
  	td->ca = virt_to_phys(skb->data);

	/* Using the NDPTR to handle the DMA Race Condition */
   	if(local_readl(&(lp->tx_dma_regs->dmandptr)) == 0) 		
  	{
		if( lp->tx_chain_status == empty ) 
		{
			td->control = DMA_COUNT(length) |DMAD_cof_m |DMAD_iof_m; 
			lp->tx_chain_tail = (lp->tx_chain_tail + 1) & ACACIA_TDS_MASK;
			local_writel(virt_to_phys(&lp->td_ring[lp->tx_chain_head]), &(lp->tx_dma_regs->dmandptr));			
			lp->tx_chain_head = lp->tx_chain_tail; 
		}
		else
		{
			lp->td_ring[(lp->tx_chain_tail-1)& ACACIA_TDS_MASK].control &=  ~(DMAD_cof_m); 	
			lp->td_ring[(lp->tx_chain_tail-1)& ACACIA_TDS_MASK].link =  virt_to_phys(td);
			td->control = DMA_COUNT(length) |DMAD_cof_m |DMAD_iof_m;
			lp->tx_chain_tail = (lp->tx_chain_tail + 1) & ACACIA_TDS_MASK;
			lp->tx_chain_status = empty;
			local_writel(virt_to_phys(&lp->td_ring[lp->tx_chain_head]), &(lp->tx_dma_regs->dmandptr));			
			lp->tx_chain_head = lp->tx_chain_tail; 
		}
		
   	}
	else
	{
		if( lp->tx_chain_status == empty )
		{
			td->control = DMA_COUNT(length) |DMAD_cof_m |DMAD_iof_m;
			lp->td_ring[(lp->tx_chain_tail-1)& ACACIA_TDS_MASK].link =  virt_to_phys(td);	
			lp->tx_chain_tail = (lp->tx_chain_tail + 1) & ACACIA_TDS_MASK;
			lp->tx_chain_status = filled;		
		}
		else
		{
			lp->td_ring[(lp->tx_chain_tail-1)& ACACIA_TDS_MASK].control &=  ~(DMAD_cof_m); 	
			lp->td_ring[(lp->tx_chain_tail-1)& ACACIA_TDS_MASK].link =  virt_to_phys(td);	
			td->control = DMA_COUNT(length) |DMAD_cof_m |DMAD_iof_m; 
			lp->tx_chain_tail = (lp->tx_chain_tail + 1) & ACACIA_TDS_MASK;
		}		
	}

	dev->trans_start = jiffies;				

     	spin_unlock_irqrestore(&lp->lock, flags);
  
  	return 0;
}

#if 0
/* Ethernet MII-PHY Handler */
static void acacia_mii_handler(unsigned long data)
{
	struct net_device *dev = (struct net_device *)data;		
	struct acacia_local *lp = (struct acacia_local *)dev->priv;
  	unsigned long 	flags;
	unsigned long duplex_status;
	int port_addr = (lp->rx_irq == 0x2c? 1:0) << 8;

	spin_lock_irqsave(&lp->lock, flags);

	/* Two ports are using the same MII, the difference is the PHY address */
	local_writel(0, &rc32438_eth0_regs->miimcfg);  
	local_writel(0, &rc32438_eth0_regs->miimcmd);  
	local_writel(port_addr |0x05, &rc32438_eth0_regs->miimaddr);  
	local_writel(MIIMCMD_scn_m, &rc32438_eth0_regs->miimcmd);  
	while(local_readl(&rc32438_eth0_regs->miimind) & MIIMIND_nv_m);

	ERR("irq:%x		port_addr:%x	RDD:%x\n", 
		lp->rx_irq, port_addr, local_readl(&rc32438_eth0_regs->miimrdd));
	duplex_status = (local_readl(&rc32438_eth0_regs->miimrdd) & 0x140)? ETHMAC2_fd_m: 0;
	if(duplex_status != lp->duplex_mode)
	{
		ERR("The MII-PHY is Auto-negotiated to %s-Duplex mode for Eth-%x\n", duplex_status? "Full":"Half", lp->rx_irq == 0x2c? 1:0);		
		lp->duplex_mode = duplex_status;
		acacia_restart(dev);		
	}

   	lp->mii_phy_timer.expires = jiffies + 10 * HZ;	
   	add_timer(&lp->mii_phy_timer);

     	spin_unlock_irqrestore(&lp->lock, flags);

}
#endif

#ifdef	RC32438_REVISION_ZA	
/* Ethernet Rx Overflow interrupt */
static void acacia_ovr_interrupt(int irq, void *dev_id, struct pt_regs * regs)
{
	struct net_device *dev = (struct net_device *)dev_id;
	struct acacia_local *lp;
	unsigned int i;
	
	ASSERT(dev != NULL);

	netif_stop_queue(dev);	
	
	lp = (struct acacia_local *)dev->priv;

	spin_lock(&lp->lock);

	i = local_readl(&lp->eth_regs->ethintfc);
  	i &= ~ETHINTFC_ovr_m;
	local_writel(i, &lp->eth_regs->ethintfc);

	/* Restart interface */
	acacia_restart(dev);     

	spin_unlock(&lp->lock);

}

#endif


/* Ethernet Tx Underflow interrupt */
static void acacia_und_interrupt(int irq, void *dev_id, struct pt_regs * regs)
{
	struct net_device *dev = (struct net_device *)dev_id;
	struct acacia_local *lp;
	unsigned int i;

	ASSERT(dev != NULL);
	
	netif_stop_queue(dev);	
	lp = (struct acacia_local *)dev->priv;

	ERR("Tx underflow - i/f reset\n");
	spin_lock(&lp->lock);

	i = local_readl(&lp->eth_regs->ethintfc);
  	i &= ~ETHINTFC_und_m;
	local_writel(i, &lp->eth_regs->ethintfc);

	/* Restart interface */
	acacia_restart(dev);     
	
	spin_unlock(&lp->lock);

}


/* Ethernet Rx DMA interrupt */
static void acacia_rx_dma_interrupt(int irq, void *dev_id, struct pt_regs * regs)
{
  	struct net_device *dev = (struct net_device *)dev_id;
	struct acacia_local* lp;
  	volatile u32 dmas;

  	ASSERT(dev != NULL);

	lp = (struct acacia_local *)dev->priv;

  	spin_lock(&lp->lock);

  	/* Mask D H E bit in Rx DMA */
	local_writel(local_readl(&lp->rx_dma_regs->dmasm) | (DMASM_d_m | DMASM_h_m | DMASM_e_m), &lp->rx_dma_regs->dmasm);

  	dmas = local_readl(&lp->rx_dma_regs->dmas);
#ifdef	RC32438_REVISION_ZA	
	local_writel(~dmas, &lp->rx_dma_regs->dmas);
#endif

	if(dmas & (DMAS_d_m | DMAS_h_m | DMAS_e_m))
    	{
      		tasklet_hi_schedule(lp->rx_tasklet);
    	}

  	spin_unlock(&lp->lock);

}



static void acacia_rx_tasklet(unsigned long rx_data_dev)
{
  struct net_device *dev = (struct net_device *)rx_data_dev;	
  struct acacia_local* lp = (struct acacia_local *)dev->priv;
  volatile DMAD_t  rd = KSEG1ADDR(&lp->rd_ring[lp->rx_next_done]);
  struct sk_buff *skb, *skb_new;
  u8* pkt_buf;
  u32 devcs, count, pkt_len, pktuncrc_len;
  unsigned long 	flags;
  volatile u32 dmas;

#ifdef CONFIG_NET_FASTROUTE
	struct ethhdr *eth;
	struct iphdr *iph;
	struct net_device *odev;
	struct rtable *rt;
	u32 CPU_ID = smp_processor_id();
	u32 fast_routed = 0;
	u32 h;
#endif				/* #ifdef CONFIG_NET_FASTROUTE */

#ifdef	RC32438_REVISION_ZA		
  struct sk_buff 	 *skb_new_1, *skb_new_2;	
#endif

  spin_lock_irqsave(&lp->lock, flags);
  
  while ( (count = ACACIA_RBSIZE - (u32)DMA_COUNT(rd->control)) != 0)
    {
      /* init the var. used for the later operations within the while loop */
#ifdef CONFIG_NET_FASTROUTE
      fast_routed = 0;
#endif				/* #ifdef CONFIG_NET_FASTROUTE */
      
      skb_new = NULL;
      devcs = rd->devcs;
      pkt_len = RCVPKT_LENGTH(devcs);
      skb = lp->rx_skb[lp->rx_next_done];
      
      if (count < 64) 
	{
	  lp->stats.rx_errors++;
	  lp->stats.rx_dropped++;			
	}
      else if ((devcs & ( ETHRX_ld_m)) !=	ETHRX_ld_m)
	{
	  /* check that this is a whole packet */
	  /* WARNING: DMA_FD bit incorrectly set in Rc32438 (errata ref #077) */
	  lp->stats.rx_errors++;
	  lp->stats.rx_dropped++;
	}
      else if ( (devcs & ETHRX_rok_m)  )
	{
	  
#ifdef	RC32438_REVISION_ZA		
	  /*
	   * Due to a bug in rc32438 ZA processor, the packet length
	   * given by devcs field and count field sometimes differ.
	   *
	   * 2 packets of lenght 0x40 bytes are merged into one packet of length
	   * 0x7c, with the last 4 bytes (CRC Fields) of the first packet
	   * being ovewwritten by the 2nd packet. 
	   *
	   * To avoid packet loss we seperate the merged packest (0x7c) and pass
	   * it up to the upper layer (netif_rx) as 2 different packets of
	   * length 0x3c and 0x40 bytes.
	   */
	  
	  if ((count == 0x7c) && ((count - pkt_len) == 60))
	    {
	      pkt_buf = (u8*)lp->rx_skb[lp->rx_next_done]->data;
	      dma_cache_inv((unsigned long)pkt_buf, count);
	      
	      skb_new_1 = dev_alloc_skb(ACACIA_RBSIZE + 2);		
	      if (skb_new_1 != NULL) 
		{
		  /* 16 bit align */
		  skb_reserve(skb_new_1, 2);			
		  skb_new_1->dev = dev;					
		  skb_put(skb_new_1, 60);
		  eth_copy_and_sum(skb_new_1, pkt_buf, 60, 0);
		  skb_new_1->protocol = eth_type_trans(skb_new_1, dev);
		  netif_rx(skb_new_1);
		}
	      else
		{
		  lp->stats.rx_errors++;
		  lp->stats.rx_dropped++;
		}
	      
	      skb_new_2 = dev_alloc_skb(ACACIA_RBSIZE + 2);				
	      if (skb_new_2 != NULL) 
		{
		  /* 16 bit align */
		  skb_reserve(skb_new_2, 2);			
		  skb_new_2->dev = dev;					
		  skb_put(skb_new_2, 60);
		  eth_copy_and_sum(skb_new_2, (pkt_buf + 60), 60, 0);
		  skb_new_2->protocol = eth_type_trans(skb_new_2, dev);
		  netif_rx(skb_new_2);
		}
	      else
		{
		  lp->stats.rx_errors++;
		  lp->stats.rx_dropped++;
		}
	      
	    }
	  else
#endif				
	    {
	      /* must be the (first and) last descriptor then */
	      pkt_buf = (u8*)lp->rx_skb[lp->rx_next_done]->data;
	      
	      pktuncrc_len = pkt_len - 4;
	      /* invalidate the cache */
	      dma_cache_inv((unsigned long)pkt_buf, pktuncrc_len);
	      
	      /* Malloc up new buffer. */					  
	      skb_new = dev_alloc_skb(ACACIA_RBSIZE + 2);					             	
	      
	      if (skb_new != NULL)
		{
		  /* Make room */
		  skb_put(skb, pktuncrc_len);		    
#ifdef CONFIG_NET_FASTROUTE
		  /*
		   * Fast Route
		   */
		  

#if 0
		  if (pkt_len > 1024)
		    goto bigpacket;
		  
#endif
		  
		  eth = (struct ethhdr *) skb->data;
		  if (eth->h_proto == __constant_htons(ETH_P_IP)) {
		    iph = (struct iphdr *) (skb->data + ETH_HLEN);
		    h = (*(u8 *) & iph->daddr
			 ^ *(u8 *) & iph->
			 saddr) & NETDEV_FASTROUTE_HMASK;
		    rt = (struct rtable *) (dev->fastpath[h]);
		    if (rt
			&& ((u16 *) & iph->daddr)[0] ==
			((u16 *) & rt->key.dst)[0]
			&& ((u16 *) & iph->daddr)[1] ==
			((u16 *) & rt->key.dst)[1]
			&& ((u16 *) & iph->saddr)[0] ==
			((u16 *) & rt->key.src)[0]
			&& ((u16 *) & iph->saddr)[1] ==
			((u16 *) & rt->key.src)[1]
			&& !rt->u.dst.obsolete) {
		      odev = rt->u.dst.dev;
		      netdev_rx_stat[CPU_ID].fastroute_hit++;
		      
		      if (*(u8 *) iph == 0x45 && !(eth->h_dest[0] & 0x80)	/* 0x80 for IP payload 0x45 for IPv4*/
			  &&neigh_is_valid(rt->u.dst.
					   neighbour)
			  && iph->ttl > 1) {	/* Fast Route Path */
			
			fast_routed = 1;
			
			if ((!netif_queue_stopped(odev)))
			  //if (1)
			  {
			    skb->pkt_type =
			      PACKET_FASTROUTE;
			    skb->protocol =
			      __constant_htons
			      (ETH_P_IP);
			    
			    ip_decrease_ttl(iph);
			    
			    
			    memcpy(eth->h_source,
				   odev->dev_addr,
				   6);
			    memcpy(eth->h_dest,
				   rt->u.dst.
				   neighbour->ha,
				   6);
			    skb->dev = odev;
			    if (odev->hard_start_xmit(skb, odev)!= 0) {
			      panic
				("I am dead");
			    }
			    netdev_rx_stat[CPU_ID].
			      fastroute_success++;
			  } else {	/* Semi Fast Route Path */
			    
			    skb->pkt_type =
			      PACKET_FASTROUTE;
			    skb->nh.raw =
			      skb->data +
			      ETH_HLEN;
			    skb->protocol =
			      __constant_htons
			      (ETH_P_IP);
			    netdev_rx_stat[CPU_ID].
			      fastroute_defer++;
			    netif_rx(skb);
			  }
		      }
		    }
		  }
		  
		  // bigpacket: 
		  
		  if (fast_routed == 0)
#endif				/* #ifdef CONFIG_NET_FASTROUTE */
		    
		    
		    {
		      skb->protocol = eth_type_trans(skb, dev);
//printk("pkt rcvd. len = %d type = 0x%x\n", pktuncrc_len, skb->protocol);
		      
		      /* pass the packet to upper layers */
		      netif_rx(skb);
		    }
		  dev->last_rx = jiffies;
		  lp->stats.rx_packets++;
		  lp->stats.rx_bytes += pktuncrc_len;
		  
		  if (IS_RCV_MP(devcs))
		    lp->stats.multicast++;
		  
		  /* 16 bit align */						  
		  skb_reserve(skb_new, 2);	
		  
		  skb_new->dev = dev;
		  lp->rx_skb[lp->rx_next_done] = skb_new;
		}
	      else
		{
		  ERR("no memory, dropping rx packet.\n");
		  lp->stats.rx_errors++;		
		  lp->stats.rx_dropped++;					
		}
	    }
	  
	}			
      else
	{
	  /* This should only happen if we enable accepting broken packets */
	  lp->stats.rx_errors++;
	  lp->stats.rx_dropped++;
	  
	  /* add statistics counters */
	  if (IS_RCV_CRC_ERR(devcs)) 
	    {
	      DBG(2, "RX CRC error\n");
	      lp->stats.rx_crc_errors++;
	    } 
	  else if (IS_RCV_LOR_ERR(devcs)) 
	    {
	      DBG(2, "RX LOR error\n");
	      lp->stats.rx_length_errors++;
	    }				
	  else if (IS_RCV_LE_ERR(devcs)) 
	    {
	      DBG(2, "RX LE error\n");
	      lp->stats.rx_length_errors++;
	    }
	  else if (IS_RCV_OVR_ERR(devcs)) 
	    {
	      /*
	       * The overflow errors are handled through
	       * an interrupt handler.
	       */
	      lp->stats.rx_over_errors++;
	    }
	  else if (IS_RCV_CV_ERR(devcs)) 
	    {		        
	      /* code violation */
	      DBG(2, "RX CV error\n");
	      lp->stats.rx_frame_errors++;
	    }
	  else if (IS_RCV_CES_ERR(devcs)) 
	    {
	      DBG(2, "RX Preamble error\n");
	    }
	}
      
      rd->devcs = 0;
      
      /* restore descriptor's curr_addr */
      if(skb_new)
	rd->ca = virt_to_phys(skb_new->data); 
      else
	rd->ca = virt_to_phys(skb->data);
		  
      rd->control = DMA_COUNT(ACACIA_RBSIZE) |DMAD_cod_m |DMAD_iod_m;
      lp->rd_ring[(lp->rx_next_done-1)& ACACIA_RDS_MASK].control &=  ~(DMAD_cod_m); 	

      lp->rx_next_done = (lp->rx_next_done + 1) & ACACIA_RDS_MASK;
      rd = KSEG1ADDR(&lp->rd_ring[lp->rx_next_done]);
      local_writel( ~DMAS_d_m, &lp->rx_dma_regs->dmas);
      
    }	
    
  dmas = local_readl(&lp->rx_dma_regs->dmas);

  if(dmas & DMAS_h_m)
    {
#ifdef ACACIA_PROC_DEBUG
      lp->dma_halt_cnt++;
#endif
      rd->devcs = 0;
      skb = lp->rx_skb[lp->rx_next_done];
      rd->ca = virt_to_phys(skb->data);
      acacia_chain_rx(lp,rd);
    }

  local_writel( ~(DMAS_h_m | DMAS_e_m), &lp->rx_dma_regs->dmas);
  
  /* Enable D H E bit in Rx DMA */
  local_writel(local_readl(&lp->rx_dma_regs->dmasm) & ~(DMASM_d_m | DMASM_h_m |DMASM_e_m), &lp->rx_dma_regs->dmasm); 
  spin_unlock_irqrestore(&lp->lock, flags);
  
}	



/* Ethernet Tx DMA interrupt */
static void acacia_tx_dma_interrupt(int irq, void *dev_id, struct pt_regs * regs)
{
	struct net_device *dev = (struct net_device *)dev_id;
	struct acacia_local *lp;
	volatile u32 dmas;

	ASSERT(dev != NULL);

	lp = (struct acacia_local *)dev->priv;
	
	spin_lock(&lp->lock);

	/* Mask F E bit in Tx DMA */
	local_writel(local_readl(&lp->tx_dma_regs->dmasm) | (DMASM_f_m | DMASM_e_m), &lp->tx_dma_regs->dmasm);

	dmas = local_readl(&lp->tx_dma_regs->dmas);

	if (dmas & (DMAS_f_m | DMAS_e_m))
	{
		tasklet_hi_schedule(lp->tx_tasklet);

		if (dmas & DMAS_e_m)
	  		ERR(__FUNCTION__ ": DMA error\n");
		
	}

	local_writel(~dmas, &lp->tx_dma_regs->dmas); 
   	if(lp->tx_chain_status == filled && (local_readl(&(lp->tx_dma_regs->dmandptr)) == 0)) 
	{
		local_writel(virt_to_phys(&lp->td_ring[lp->tx_chain_head]), &(lp->tx_dma_regs->dmandptr));			
		lp->tx_chain_status = empty;
		lp->tx_chain_head = lp->tx_chain_tail;
	}

	spin_unlock(&lp->lock);
}


static void acacia_tx_tasklet(unsigned long tx_data_dev)
{
	struct net_device *dev = (struct net_device *)tx_data_dev;	
  	struct acacia_local* lp = (struct acacia_local *)dev->priv;
  	volatile DMAD_t td = KSEG1ADDR(&lp->td_ring[lp->tx_next_done]);
	u32 devcs;
    	unsigned long 	flags;
	
	spin_lock_irqsave(&lp->lock, flags);

  	/* process all desc that are done */
  	while(IS_DMA_FINISHED(td->control)) 
  	{
		if(lp->tx_full == 1)
		{
			netif_wake_queue(dev);
			lp->tx_full = 0;
		}

		devcs = lp->td_ring[lp->tx_next_done].devcs;    
		if ((devcs & (ETHTX_fd_m | ETHTX_ld_m)) != (ETHTX_fd_m | ETHTX_ld_m)) 
		{
	      		lp->stats.tx_errors++;
	      		lp->stats.tx_dropped++;				

	      		/* should never happen */
	      		DBG(1, __FUNCTION__ ": split tx ignored\n");
	    	}
		else if (IS_TX_TOK(devcs)) 
		{
	      		/* transmit OK */
	      		lp->stats.tx_packets++;
			lp->stats.tx_bytes += lp->tx_skb[lp->tx_next_done]->len;
	    	}
		else 
		{
	      		lp->stats.tx_errors++;
	      		lp->stats.tx_dropped++;				
			  
		      	/* underflow */
		  	if (IS_TX_UND_ERR(devcs)) 
				lp->stats.tx_fifo_errors++;

		      	/* oversized frame */
		      	if (IS_TX_OF_ERR(devcs))
				lp->stats.tx_aborted_errors++;

		      	/* excessive deferrals */
		      	if (IS_TX_ED_ERR(devcs))
				lp->stats.tx_carrier_errors++;

		      	/* collisions: medium busy */
		      	if (IS_TX_EC_ERR(devcs))
				lp->stats.collisions++;

		      	/* late collision */
		      	if (IS_TX_LC_ERR(devcs))
				lp->stats.tx_window_errors++;

    		}

	    	/* We must always free the original skb */
	    	if (lp->tx_skb[lp->tx_next_done] != NULL) 
		{
	      		dev_kfree_skb_any(lp->tx_skb[lp->tx_next_done]);
	      		lp->tx_skb[lp->tx_next_done] = NULL;
	    	}

		lp->td_ring[lp->tx_next_done].control = DMAD_iof_m;
		lp->td_ring[lp->tx_next_done].devcs = ETHTX_fd_m | ETHTX_ld_m;	
		lp->td_ring[lp->tx_next_done].link = 0;
		lp->td_ring[lp->tx_next_done].ca = 0;
	    	lp->tx_count --;

	    	/* go on to next transmission */
	    	lp->tx_next_done = (lp->tx_next_done + 1) & ACACIA_TDS_MASK;
	    	td = KSEG1ADDR(&lp->td_ring[lp->tx_next_done]);
  	}


/*	
	dmas = local_readl(&lp->tx_dma_regs->dmas);
	if(dmas & DMAS_e_m)
	{
		ERR(__FUNCTION__ ": DMAS_e_m\n");
	    	lp->stats.tx_errors++;	
	}
	local_writel( ~dmas, &lp->tx_dma_regs->dmas);
*/

  	/* Enable F E bit in Tx DMA */
  	local_writel(local_readl(&lp->tx_dma_regs->dmasm) & ~(DMASM_f_m | DMASM_e_m), &lp->tx_dma_regs->dmasm); 
   	spin_unlock_irqrestore(&lp->lock, flags);
	
}

/*
 * Get the current statistics.
 * This may be called with the device open or closed.
 */
static struct net_device_stats * acacia_get_stats(struct net_device *dev)
{
	struct acacia_local *lp = (struct acacia_local *)dev->priv;
	return &lp->stats;
}


/*
 * Set or clear the multicast filter for this adaptor.
 */
static void acacia_multicast_list(struct net_device *dev)
{   
	/* listen to broadcasts always and to treat 	*/
	/*       IFF bits independantly	*/
	struct acacia_local *lp = (struct acacia_local *)dev->priv;
	unsigned long flags;
	u32 recognise = ETHARC_ab_m; 		/* always accept broadcasts */

	if (dev->flags & IFF_PROMISC)         		/* set promiscuous mode */
		recognise |= ETHARC_pro_m;

	if ((dev->flags & IFF_ALLMULTI) || (dev->mc_count > 15))
		recognise |= ETHARC_am_m;    	  	/* all multicast & bcast */
	else if (dev->mc_count > 0) 
	{
		DBG(2, __FUNCTION__ ": mc_count %d\n", dev->mc_count);
		recognise |= ETHARC_am_m;    	  	/* for the time being */
	}
	
	spin_lock_irqsave(&lp->lock, flags);
	local_writel(recognise, &lp->eth_regs->etharc);
	spin_unlock_irqrestore(&lp->lock, flags);
}


static void acacia_tx_timeout(struct net_device *dev)
{
	struct acacia_local *lp = (struct acacia_local *)dev->priv;
	unsigned long flags;

	spin_lock_irqsave(&lp->lock, flags);
	acacia_restart(dev);
	spin_unlock_irqrestore(&lp->lock, flags);
	
}


/*
 * Initialize the ACACIA ethernet controller.
 */
static int acacia_init(struct net_device *dev)
{
	struct acacia_local *lp = (struct acacia_local *)dev->priv;
	int i, j;

	/* Disable DMA */       
	acacia_abort_tx(dev);
	acacia_abort_rx(dev); 
	
	/* reset ethernet logic */ 
	local_writel(0, &lp->eth_regs->ethintfc);
	while((local_readl(&lp->eth_regs->ethintfc) & ETHINTFC_rip_m))
		dev->trans_start = jiffies;	
	
	/* Enable Ethernet Interface */ 
	local_writel(ETHINTFC_en_m, &lp->eth_regs->ethintfc); 

	tasklet_disable(lp->rx_tasklet);
	tasklet_disable(lp->tx_tasklet);

	/* Initialize the transmit Descriptors */
	for (i = 0; i < ACACIA_NUM_TDS; i++) 
	{
    		lp->td_ring[i].control = DMAD_iof_m;
    		lp->td_ring[i].devcs = ETHTX_fd_m | ETHTX_ld_m;
    		lp->td_ring[i].ca = 0;
    		lp->td_ring[i].link = 0;
    		if (lp->tx_skb[i] != NULL) 
		{
      			/* free dangling skb */
      			dev_kfree_skb_any(lp->tx_skb[i]);
      			lp->tx_skb[i] = NULL;
    		}
	}
    	lp->tx_next_done = lp->tx_chain_head = lp->tx_chain_tail = 	lp->tx_full = lp->tx_count = 0;
	lp->	tx_chain_status = empty;
	
	/*
	 * Initialize the receive descriptors so that they
	 * become a circular linked list, ie. let the last
	 * descriptor point to the first again.
	 */
  	for (i=0; i<ACACIA_NUM_RDS; i++) 
  	{
		struct sk_buff *skb = lp->rx_skb[i];

		if (lp->rx_skb[i] == NULL) 
		{
			skb = dev_alloc_skb(ACACIA_RBSIZE + 2);
			if (skb == NULL)
			{
		 		ERR("No memory in the system\n");
				for (j = 0; j < ACACIA_NUM_RDS; j ++)
					if (lp->rx_skb[j] != NULL) 
						dev_kfree_skb_any(lp->rx_skb[j]);
				
	              	return 1;
			}
			else
			{
				skb->dev = dev;
				skb_reserve(skb, 2);
				lp->rx_skb[i] = skb;
		             	lp->rd_ring[i].ca = virt_to_phys(skb->data); 
				
			}
		}
		lp->rd_ring[i].control =	DMAD_iod_m | DMA_COUNT(ACACIA_RBSIZE);
		lp->rd_ring[i].devcs = 0;
              lp->rd_ring[i].ca = virt_to_phys(skb->data);
		lp->rd_ring[i].link = virt_to_phys(&lp->rd_ring[i+1]);

  	}
    	/* loop back */
  	lp->rd_ring[ACACIA_NUM_RDS-1].link = virt_to_phys(&lp->rd_ring[0]);
    	lp->rx_next_done   = 0;

  	lp->rd_ring[ACACIA_NUM_RDS-1].control |= DMAD_cod_m;
  	lp->rx_chain_head = 0;
  	lp->rx_chain_tail = 0;
  	lp->rx_chain_status = empty;

	local_writel(0, &lp->rx_dma_regs->dmas);
	/* Start Rx DMA */
	acacia_start_rx(lp, &lp->rd_ring[0]);
	
  	/* Enable F E bit in Tx DMA */
  	local_writel(local_readl(&lp->tx_dma_regs->dmasm) & ~(DMASM_f_m | DMASM_e_m), &lp->tx_dma_regs->dmasm); 
    	/* Enable D H E bit in Rx DMA */
  	local_writel(local_readl(&lp->rx_dma_regs->dmasm) & ~(DMASM_d_m | DMASM_h_m | DMASM_e_m), &lp->rx_dma_regs->dmasm); 

	/* Accept only packets destined for this Ethernet device address */
	local_writel(ETHARC_ab_m, &lp->eth_regs->etharc); 
 
	/* Set all Ether station address registers to their initial values */ 
	local_writel(STATION_ADDRESS_LOW(dev), &lp->eth_regs->ethsal0); 
	local_writel(STATION_ADDRESS_HIGH(dev), &lp->eth_regs->ethsah0);
	
	local_writel(STATION_ADDRESS_LOW(dev), &lp->eth_regs->ethsal1); 
	local_writel(STATION_ADDRESS_HIGH(dev), &lp->eth_regs->ethsah1);
	
	local_writel(STATION_ADDRESS_LOW(dev), &lp->eth_regs->ethsal2); 
	local_writel(STATION_ADDRESS_HIGH(dev), &lp->eth_regs->ethsah2);
	
	local_writel(STATION_ADDRESS_LOW(dev), &lp->eth_regs->ethsal3); 
	local_writel(STATION_ADDRESS_HIGH(dev), &lp->eth_regs->ethsah3); 
 

	/* Frame Length Checking, Pad Enable, CRC Enable, Full Duplex set */ 
	local_writel(ETHMAC2_pe_m | ETHMAC2_cen_m | ETHMAC2_fd_m, &lp->eth_regs->ethmac2);  
	//ETHMAC2_flc_m		ETHMAC2_fd_m	lp->duplex_mode

	/* Back to back inter-packet-gap */ 
	local_writel(0x15, &lp->eth_regs->ethipgt); 
	/* Non - Back to back inter-packet-gap */ 
	local_writel(0x12, &lp->eth_regs->ethipgr); 
     
	/* Management Clock Prescaler Divisor */
	/* Clock independent setting */
	local_writel(((idt_cpu_freq)/MII_CLOCK+1) & ~1,
	       &lp->eth_regs->ethmcp);

	/* don't transmit until fifo contains 48b */
	local_writel(48, &lp->eth_regs->ethfifott);

	local_writel(ETHMAC1_re_m, &lp->eth_regs->ethmac1);

	tasklet_enable(lp->rx_tasklet);
	tasklet_enable(lp->tx_tasklet);

	netif_start_queue(dev);


	return 0; 

}


static void acacia_cleanup_module(void)
{
	int i;

	for (i = 0; acacia_iflist[i].iobase; i++) 
	{
		struct acacia_if_t * bif = &acacia_iflist[i];
		if (bif->dev != NULL) 
		{
			struct acacia_local *lp = (struct acacia_local *)bif->dev->priv;
			if (lp != NULL) 
			{
				if (lp->eth_regs)
					iounmap((void*)lp->eth_regs);
				if (lp->rx_dma_regs)
					iounmap((void*)lp->rx_dma_regs);
				if (lp->tx_dma_regs)
					iounmap((void*)lp->tx_dma_regs);
				if (lp->td_ring)
					kfree((void*)KSEG0ADDR(lp->td_ring));

#ifdef ACACIA_PROC_DEBUG
				if (lp->ps)
					remove_proc_entry("net/rc32438", NULL);
#endif
				kfree(lp);
			}

			unregister_netdev(bif->dev);
			kfree(bif->dev);
			release_region(bif->iobase, 0x24C);
		}
	}
}


#ifndef MODULE

static int __init acacia_setup(char *options)
{
	/* no options yet */
	return 1;
}

static int __init acacia_setup_ethaddr0(char *options)
{
	memcpy(mac0, options, 6);
	mac0[6]= '\0';
	return 1;
}

static int __init acacia_setup_ethaddr1(char *options)
{
	memcpy(mac1, options, 6);
	mac1[6]= '\0';
	return 1;
}

__setup("rc32438eth=", acacia_setup);
__setup("ethaddr0=", acacia_setup_ethaddr0);
__setup("ethaddr1=", acacia_setup_ethaddr1);


#endif /* MODULE */

module_init(acacia_init_module);
module_exit(acacia_cleanup_module);














